import pytest

from whispers.plugins import WhisperPlugins
from whispers.plugins.config import Config
from whispers.plugins.dockerfile import Dockerfile
from whispers.plugins.htpasswd import Htpasswd
from whispers.plugins.jproperties import Jproperties
from whispers.plugins.json import Json
from whispers.plugins.npmrc import Npmrc
from whispers.plugins.pip import Pip
from whispers.plugins.plaintext import Plaintext
from whispers.plugins.pypirc import Pypirc
from whispers.plugins.python import Python
from whispers.plugins.shell import Shell
from whispers.plugins.xml import Xml
from whispers.plugins.yml import Yml

from .conftest import fixture_path


@pytest.mark.parametrize(
    ("filename", "expected_plugin"),
    [
        ("File.404", type(None)),
        ("", type(None)),
        (".aws/credentials", Config),
        (".htpasswd", Htpasswd),
        (".npmrc", Npmrc),
        (".pypirc", Pypirc),
        ("apikeys.json", Json),
        ("apikeys.xml", Xml),
        ("apikeys.yml", Yml),
        ("connection.config", Xml),
        ("cors.py", Python),
        ("Dockerfile", Dockerfile),
        ("integration.conf", Xml),
        ("java.properties", Jproperties),
        ("pip.conf", Pip),
        ("plaintext.txt", Plaintext),
        ("settings.cfg", Config),
        ("settings.env", Config),
        ("settings.ini", Config),
        ("shell.sh", Shell),
    ],
)
def test_init(filename, expected_plugin):
    filename = fixture_path(filename)
    plugin = WhisperPlugins(filename).plugin
    assert isinstance(plugin, expected_plugin)



from pathlib import Path

from lxml import etree as ElementTree

from whispers.log import debug
from whispers.plugins.uri import Uri
from whispers.rules import WhisperRules


class Xml:
    def __init__(self):
        self.breadcrumbs = []
        self.rules = WhisperRules()

    def pairs(self, filepath: Path):
        def _traverse(tree):
            """Traverse XML document"""
            for event, element in tree:
                if event == "start":
                    self.breadcrumbs.append(element.tag)
                elif event == "end":
                    self.breadcrumbs.pop()
                    continue

                # Format: <elem key="value">
                for key, value in element.attrib.items():
                    yield key, value, self.breadcrumbs

                    # Format: <elem name="jdbc:mysql://host?k1=v1&amp;k2=v2">
                    if self.rules.match("uri", value):
                        for k, v in Uri().pairs(value):
                            yield k, v, self.breadcrumbs

                # Format: <key>value</key>
                if not element.text:
                    continue
                yield element.tag, element.text, self.breadcrumbs

                # Format: <elem>key=value</elem>
                if "=" in element.text:
                    item = element.text.split("=")
                    if len(item) == 2:
                        yield item[0], item[1], self.breadcrumbs

                # Format: <key>name</key><value>string</value>
                found_key = None
                found_value = None
                for item in element:
                    if str(item.tag).lower() == "key":
                        found_key = item.text
                    elif str(item.tag).lower() == "value":
                        found_value = item.text
                if found_key and found_value:
                    yield found_key, found_value, self.breadcrumbs

        try:
            parser = ElementTree.XMLParser(recover=True)
            tree = ElementTree.parse(filepath.as_posix(), parser)
            tree = ElementTree.iterwalk(tree, events=("start", "end"))
            yield from _traverse(tree)
        except Exception as e:
            debug(f"{type(e)} in {filepath}")



from argparse import ArgumentParser
from pathlib import Path

from whispers.__version__ import __version__
from whispers.core import load_config, run
from whispers.log import configure_log
from whispers.utils import format_stdout


def cli():
    # Parse CLI arguments
    args_parser = ArgumentParser("whispers", description=("Identify secrets and dangerous behaviours"))
    args_parser.add_argument("-v", "--version", action="version", version=f"whispers {__version__}")
    args_parser.add_argument("-c", "--config", default=None, help="config file")
    args_parser.add_argument("-o", "--output", help="output file (.yml)")
    args_parser.add_argument("src", nargs="?", help="source code file or directory")
    args = args_parser.parse_args()

    # Default response
    if not args.src:
        exit(args_parser.print_help())

    # Clear output file
    if args.output:
        args.output = Path(args.output)
        args.output.write_text("")

    # Configure execution
    configure_log()
    if args.config:
        args.config = load_config(args.config, src=args.src)

    # Valar margulis
    for secret in run(args.src, config=args.config):
        format_stdout(secret, args.output)


if __name__ == "__main__":
    cli()



from pathlib import Path


class Pypirc:
    def pairs(self, filepath: Path):
        for line in filepath.open("r").readlines():
            if "password:" not in line:
                continue

            value = line.split("password:")[-1].strip()
            if value:
                yield "PyPI password", value



from pathlib import Path

import pytest

from whispers.rules import WhisperRules
from whispers.utils import load_yaml_from_file

from .conftest import does_not_raise, rule_path


@pytest.mark.parametrize(
    ("rulefile", "expectation"),
    [
        ("empty.yml", does_not_raise()),
        ("valid.yml", does_not_raise()),
        ("multiple.yml", does_not_raise()),
        ("invalid_severity.yml", pytest.raises(ValueError)),
    ],
)
def test_load_rules(rulefile, expectation):
    rules = WhisperRules(rule_path("empty.yml"))
    rulefile = rule_path(rulefile)
    ruleyaml = load_yaml_from_file(Path(rulefile))
    with expectation:
        rules.load_rules(rulefile)
        assert len(rules.rules) == len(ruleyaml)
        for rule_id in ruleyaml.keys():
            assert rule_id in rules.rules


@pytest.mark.parametrize(
    ("rulefile", "expectation"),
    [("valid.yml", does_not_raise()), ("doesnotexist.yml", pytest.raises(FileNotFoundError))],
)
def test_load_rules_from_file(rulefile, expectation):
    rules = WhisperRules()
    rules_len = len(rules.rules)
    with expectation:
        rules.load_rules_from_file(Path(rule_path(rulefile)))
        assert len(rules.rules) == rules_len + 1


@pytest.mark.parametrize(
    ("rulefile", "rules_added"), [("empty.yml", 0), ("valid.yml", 1), ("multiple.yml", 4)],
)
def test_load_rules_from_dict(rulefile, rules_added):
    rules = WhisperRules()
    rules_len = len(rules.rules)
    custom_rules = load_yaml_from_file(Path(rule_path(rulefile)))
    rules.load_rules_from_dict(custom_rules)
    assert len(rules.rules) == rules_len + rules_added


@pytest.mark.parametrize(("dups", "expectation"), [(1, does_not_raise()), (2, pytest.raises(IndexError))])
def test_load_rule(dups, expectation):
    rules = WhisperRules()
    rulefile = Path(rule_path("valid.yml"))
    rule_id, rule = load_yaml_from_file(rulefile).popitem()
    with expectation:
        for _ in range(dups):
            rules.load_rule(rule_id, rule)
            assert rule_id in rules.rules
            assert rules.rules[rule_id] == rule


@pytest.mark.parametrize(
    ("rulefile", "expectation"), [("invalid_severity.yml", pytest.raises(ValueError))],
)
def test_parse_rule(rulefile, expectation):
    rules = WhisperRules()
    rulefile = Path(rule_path(rulefile))
    rule_id, rule = load_yaml_from_file(rulefile).popitem()
    with expectation:
        parsed_rule = rules.parse_rule(rule_id, rule)
        assert parsed_rule["message"] == rule["message"]
        # TODO: add the rest of asserts


@pytest.mark.parametrize(("value", "result"), [("test", True), ("Test", False), ("1test", False)])
def test_match(value, result):
    rules = WhisperRules(rule_path("valid.yml"))
    assert rules.match("valid", value) == result



from pathlib import Path

from jproperties import Properties


class Jproperties:
    def pairs(self, filepath: Path):
        props = Properties()
        props.load(filepath.read_text(), "utf-8")
        for key, value in props.properties.items():
            yield key, value



from contextlib import contextmanager
from pathlib import Path

FIXTURE_PATH = Path("tests/fixtures")
CONFIG_PATH = Path("tests/configs")
RULE_PATH = Path("tests/rules")


@contextmanager
def does_not_raise():
    yield


def fixture_path(filename: str) -> str:
    return FIXTURE_PATH.joinpath(filename).as_posix()


def config_path(filename: str) -> str:
    return CONFIG_PATH.joinpath(filename).as_posix()


def rule_path(filename: str) -> str:
    return RULE_PATH.joinpath(filename).as_posix()



import re
from base64 import b64decode
from pathlib import Path

from whispers.utils import Secret, find_line_number, load_yaml_from_file, similar_strings


class WhisperRules:
    def __init__(self, rulespath: str = ""):
        self.rules = {}
        self.load_rules(rulespath)

    def load_rules(self, rulespath: str = ""):
        if not rulespath:
            rulespath = Path(__file__).parent
        else:
            rulespath = Path(rulespath)
        if rulespath.is_dir():
            for rulefile in rulespath.glob("*.yml"):
                self.load_rules_from_file(rulefile)
        elif rulespath.is_file():
            self.load_rules_from_file(rulespath)
        else:
            raise TypeError("Rules must be loaded from a file or directory")

    def load_rules_from_file(self, rulefile: Path):
        if not rulefile.exists():
            raise FileNotFoundError(f"Rule file {rulefile.as_posix()} not found")
        rules = load_yaml_from_file(rulefile)
        for rule_id, rule in rules.items():
            self.load_rule(rule_id, rule)

    def load_rules_from_dict(self, custom_rules: dict):
        if not custom_rules:
            return
        for rule_id, rule in custom_rules.items():
            self.load_rule(rule_id, rule)

    def load_rule(self, rule_id: str, rule: dict):
        if rule_id in self.rules:
            raise IndexError(f"Duplicated rule {rule_id}, {self.rules[rule_id]}")
        self.rules[rule_id] = self.parse_rule(rule_id, rule)

    @staticmethod
    def parse_rule(rule_id: str, rule: dict) -> dict:
        required_severity = ["BLOCKER", "CRITICAL", "MAJOR", "MINOR", "INFO"]
        if rule["severity"] not in required_severity:
            raise ValueError(f"Missing severity specification in rule {rule_id}")
        for idx in ("key", "value"):
            if idx not in rule:
                continue
            if "regex" not in rule[idx]:
                continue
            flags = 0
            if rule[idx]["ignorecase"]:
                flags = re.IGNORECASE
            rule[idx]["regex"] = re.compile(rule[idx]["regex"], flags=flags)
        return rule

    def match(self, rule_id: str, text: str):
        for idx in ("key", "value"):
            if idx in self.rules[rule_id]:
                regex = self.rules[rule_id][idx]["regex"]
                if regex.match(text):
                    return True
        return False

    def check(self, key, value, filepath):
        def decode_if_base64(mkey, mvalue):
            if "isBase64" in mkey:
                if mkey["isBase64"]:
                    mvalue = b64decode(mvalue)
            return mvalue

        def is_ascii(data):
            if not isinstance(data, str):
                return False
            for ch in data:
                if ord(ch) not in range(32, 127):
                    return False
            return True

        def check_minlen(rule, mkey, mvalue):
            return len(mvalue) >= int(rule[mkey]["minlen"])

        def check_regex(rule, mkey, mvalue):
            return rule[mkey]["regex"].match(mvalue)

        def check_isBase64(rule, mkey, mvalue):
            return rule[mkey]["isBase64"] == self.match("base64", mvalue)

        def check_isAscii(rule, mkey, mvalue):
            mvalue = decode_if_base64(mkey, mvalue)
            return is_ascii(mvalue)

        def check_isUri(rule, mkey, mvalue):
            mvalue = decode_if_base64(mkey, mvalue)
            return rule[mkey]["isUri"] == self.match("uri", mvalue)

        def check_similar(rule, key, value):
            return similar_strings(key, value) < rule["similar"]

        matrix = {"key": key, "value": value}
        checks = {
            "minlen": check_minlen,
            "regex": check_regex,
            "isBase64": check_isBase64,
            "isAscii": check_isAscii,
            "isUri": check_isUri,
        }
        for rule_id, rule in self.rules.items():
            rule_matched = True
            if rule["severity"] == "INFO":
                continue
            if "similar" in rule:
                if similar_strings(key, value) >= rule["similar"]:
                    rule_matched = False
            for check_idx, check_function in checks.items():
                if not rule_matched:
                    break
                for mkey, mvalue in matrix.items():
                    if mkey not in rule:
                        continue
                    if check_idx not in rule[mkey]:
                        continue
                    if not check_function(rule, mkey, mvalue):
                        rule_matched = False
                        break
            if not rule_matched:
                continue
            return Secret(
                filepath.as_posix(),
                find_line_number(filepath, key, value),
                key,
                value,
                self.rules[rule_id]["message"],
                self.rules[rule_id]["severity"],
            )



from pathlib import Path

import astroid

from whispers.log import debug


class Python:
    def pairs(self, filepath: Path):
        """
        Parses Python source code into AST using Astroid
        http://pylint.pycqa.org/projects/astroid/en/latest/api/astroid.nodes.html#id1
        Returns key-value pairs
        """

        def is_key(node):
            types = [astroid.node_classes.Name, astroid.node_classes.AssignName]
            return type(node) in types

        def is_value(node):
            types = [astroid.node_classes.Const]
            return type(node) in types

        def _traverse(tree):
            for node in tree.get_children():
                yield from _traverse(node)
                # Assignment
                if isinstance(node, astroid.node_classes.Assign):
                    if not is_value(node.value):
                        continue
                    value = node.value.value
                    for key in node.targets:
                        if is_key(key):
                            yield key.name, value
                # Comparison
                elif isinstance(node, astroid.node_classes.Compare):
                    left = node.left
                    right = node.ops[0][1]
                    if is_key(left) and is_value(right):
                        key = left.name
                        value = right.value
                    elif is_key(right) and is_value(left):
                        key = right.name
                        value = left.value
                    else:
                        continue
                    yield key, value
                # Dictionary values
                elif isinstance(node, astroid.node_classes.Dict):
                    for key, value in node.items:
                        if not is_value(key) or not is_value(value):
                            continue
                        yield key.value, value.value
                # Keywords
                elif isinstance(node, astroid.node_classes.Keyword):
                    if is_value(node.value):
                        yield node.arg, node.value.value
                # Function call
                elif isinstance(node, astroid.node_classes.Call):
                    key = "function"
                    value = node.as_string()  # Entire function call
                    yield key, value

        try:
            tree = astroid.parse(filepath.read_text())
            yield from _traverse(tree)
        except Exception as e:
            debug(f"{type(e)} in {filepath}")



from pathlib import Path


class Npmrc:
    def pairs(self, filepath: Path):
        for line in filepath.open("r").readlines():
            if ":_authToken=" not in line:
                continue

            value = line.split(":_authToken=")[-1].strip()
            if value:
                yield "npm authToken", value



import re
from pathlib import Path

import yaml

from whispers.log import debug
from whispers.plugins.traverse import StructuredDocument


class Yml(StructuredDocument):
    def pairs(self, filepath: Path):
        def _constructor(loader, tag_suffix, node):
            """This is needed to parse IaC syntax"""
            if isinstance(node, yaml.MappingNode):
                return loader.construct_mapping(node)
            if isinstance(node, yaml.SequenceNode):
                return loader.construct_sequence(node)
            ret = loader.construct_scalar(node)
            return f"{tag_suffix} {ret}"

        """
        Convert custom YAML to parsable YAML
        - Skip ---
        - Quote unquoted values such as {{ placeholder }}
        - Remove text between <% %> and {% %}
        - Remove comments that start with #
        """
        document = ""
        for line in filepath.open("r").readlines():
            if line.startswith("---"):
                continue
            if re.match(r".+(\[)?\{\{.*\}\}(\])?", line):
                line = line.replace('"', "'")
                line = line.replace("{{", '"{{').replace("}}", '}}"')
            document += line
        document = re.sub(r"[<{]%.*?%[}>]", "", document, flags=re.MULTILINE | re.DOTALL)
        document = re.sub(r"^#.*$", "", document)

        # Load converted YAML
        yaml.add_multi_constructor("", _constructor, Loader=yaml.SafeLoader)
        try:
            code = yaml.safe_load(document)
            yield from self.traverse(code)
        except Exception as e:
            debug(f"{type(e)} in {filepath}")



import json
from pathlib import Path

import pytest

from whispers import core
from whispers.utils import (
    Secret,
    format_secret,
    format_stdout,
    line_begins_with_value,
    line_with_key_value,
    line_with_value,
    load_yaml_from_file,
    secret_checksum,
    similar_strings,
    simple_string,
    strip_string,
)

from .conftest import fixture_path, rule_path


@pytest.mark.parametrize(
    "rawstr",
    [
        "'whispers'",
        "\"'whispers'\"",
        "''whispers\"",
        "whispers\n\n",
        "\t\twhispers",
        "whispers\r\n",
        '    \t\'whispers""\r\n',
    ],
)
def test_strip_string(rawstr):
    assert strip_string(rawstr) == "whispers"


def test_simple_string():
    assert simple_string("~|wHisP3R5~|") == "__whisp3r5__"


@pytest.mark.parametrize(
    ("str1", "str2"),
    [
        ("whispers", "WHISPERS!!!"),
        ("W h  I S P e r s", "WHISPERS!!!"),
        pytest.param("AAAAAAA", "WHISPERS!!!", marks=pytest.mark.xfail),
    ],
)
def test_similar_strings(str1, str2):
    assert similar_strings(str1, str2)


@pytest.mark.parametrize(
    ("key", "value", "line"),
    [
        ("key", "value", "key=value"),
        ("key", "value", "{'key': 'value'}"),
        ("key", "value", "key -> value"),
        pytest.param("key", "value", "key=", marks=pytest.mark.xfail),
        pytest.param("key", "value", "{'key':", marks=pytest.mark.xfail),
        pytest.param("key", "value", "key ->", marks=pytest.mark.xfail),
    ],
)
def test_line_with_key_value(key, value, line):
    assert line_with_key_value(key, value, line)


@pytest.mark.parametrize(
    ("value", "line"),
    [
        ("value", "key=value"),
        ("value", "{'key': 'value'}"),
        ("value", "key -> value"),
        pytest.param("value", "key", marks=pytest.mark.xfail),
        pytest.param("value", "vaLUe", marks=pytest.mark.xfail),
    ],
)
def test_line_with_value(value, line):
    assert line_with_value(value, line)


@pytest.mark.parametrize(
    ("value", "line"),
    [
        ("value", "Value"),
        ("value", "'value'"),
        ("value", "\t\tvalue"),
        pytest.param("value", "---Value", marks=pytest.mark.xfail),
        pytest.param("value", "key=value", marks=pytest.mark.xfail),
        pytest.param("value", "\r\nvalue-2", marks=pytest.mark.xfail),
    ],
)
def test_line_begins_with_value(value, line):
    assert line_begins_with_value(value, line)


@pytest.mark.parametrize(
    ("src", "linenumbers"), [("hardcoded.yml", [12, 14, 15, 16, 19]), ("privatekeys.yml", [5, 7, 11, 12, 13, 14])],
)
def test_find_line_number(src, linenumbers):
    secrets = core.run(fixture_path(src))
    for number in linenumbers:
        assert next(secrets).line == number


@pytest.mark.parametrize(
    ("rulefile", "expected_count"), [("empty.yml", 0), ("valid.yml", 1), ("multiple.yml", 4), ("invalid.yml", 0)],
)
def test_load_yaml_from_file(rulefile, expected_count):
    rulefile = Path(rule_path(rulefile))
    assert len(load_yaml_from_file(rulefile)) == expected_count


def test_secret_checksum():
    secret = Secret("file", 123, "key", "value", "message", "severity")
    assert secret_checksum(secret) == "6370fb4455c053420588d92bd292d371"


def test_format_secret():
    secret = Secret("file", 123, "key", "value", "message", "severity")
    assert format_secret(secret) == (
        "6370fb4455c053420588d92bd292d371:\n  "
        + 'file: "file"\n  line: "123"\n  key: "key"\n  '
        + 'value: "value"\n  message: "message"\n  severity: "severity"\n\n'
    )


def test_format_stdout():
    secret = Secret("file", 123, "key", "value", "message", "severity")
    secret_str = json.dumps(secret._asdict())
    assert format_stdout(secret) == secret_str



import re
from os import urandom
from pathlib import Path

import pytest
from yaml.parser import ParserError

from whispers import core

from .conftest import FIXTURE_PATH, config_path, fixture_path


@pytest.mark.parametrize(
    ("filename", "exception"), [(f"/tmp/File404-{urandom(30).hex()}", FileNotFoundError), ("/dev/null", TypeError)],
)
def test_core_exception(filename, exception):
    with pytest.raises(exception):
        next(core.run(filename))


def test_load_config_exception():
    with pytest.raises(ParserError):
        core.load_config(config_path("invalid.yml"), FIXTURE_PATH)


def test_load_config():
    config = core.load_config(config_path("example.yml"), FIXTURE_PATH)
    assert set(config["exclude"]["files"]) == set(
        [
            Path(fixture_path(".npmrc")),
            Path(fixture_path("hardcoded.json")),
            Path(fixture_path("hardcoded.yml")),
            Path(fixture_path("hardcoded.xml")),
        ]
    )
    assert config["exclude"]["keys"] == [re.compile("SECRET_VALUE_KEY", flags=re.IGNORECASE)]
    assert config["exclude"]["values"] == [re.compile("SECRET_VALUE_PLACEHOLDER", flags=re.IGNORECASE)]


def test_include_files():
    config = core.load_config(config_path("include_files.yml"), FIXTURE_PATH)
    secrets = core.run(FIXTURE_PATH, config=config)
    assert next(secrets).value == "hardcoded"
    with pytest.raises(StopIteration):
        next(secrets)


def test_exclude_files():
    config = core.load_config(config_path("exclude_files.yml"), FIXTURE_PATH)
    secrets = core.run(FIXTURE_PATH, config=config)
    with pytest.raises(StopIteration):
        next(secrets)



from pathlib import Path

from whispers.plugins.uri import Uri
from whispers.rules import WhisperRules
from whispers.utils import strip_string


class Plaintext:
    def __init__(self):
        self.rules = WhisperRules()

    def pairs(self, filepath: Path):
        lines = filepath.open("r").readlines()
        for idx in range(len(lines)):
            line = lines[idx]
            if not strip_string(line):
                continue

            for value in line.split():
                if self.rules.match("uri", value):
                    yield from Uri().pairs(value)



from pathlib import Path


class Config:
    def pairs(self, filepath: Path):
        for line in filepath.open("r").readlines():
            if "=" not in line:
                continue
            if line.index("=") == len(line) - 1:
                continue

            item = line.strip().split("=", 1)
            yield item[0], item[1]



import re
from pathlib import Path

from whispers.log import debug
from whispers.secrets import WhisperSecrets
from whispers.utils import load_yaml_from_file


def load_config(configfile, src="."):
    configfile = Path(configfile)
    if not configfile.exists():
        debug(f"{configfile} does not exist")
        raise FileNotFoundError

    if not configfile.is_file():
        debug(f"{configfile} is not a file")
        raise TypeError

    config = load_yaml_from_file(configfile)

    # Ensure minimal expected config structure
    if "exclude" not in config:
        config["exclude"] = {"files": [], "keys": [], "values": []}
    else:
        for idx in ["files", "keys", "values"]:
            if idx not in config["exclude"]:
                config["exclude"][idx] = []
    if "include" not in config:
        config["include"] = {"files": ["**/*"]}
    elif "files" not in config["include"]:
        config["include"]["files"] = ["**/*"]

    # Glob excluded files
    exfiles = []
    for fileglob in config["exclude"]["files"]:
        for filepath in Path(src).glob(fileglob):
            exfiles.append(filepath)
    config["exclude"]["files"] = exfiles

    # Compile regex from excluded keys and values
    for param in ["keys", "values"]:
        excluded = []
        for item in config["exclude"][param]:
            excluded.append(re.compile(item, flags=re.IGNORECASE))
        config["exclude"][param] = excluded

    # Optional: rules
    if "rules" not in config:
        config["rules"] = {}

    return config


def run(src: str, config=None):
    src = Path(src)
    if not src.exists():
        debug(f"{src} does not exist")
        raise FileNotFoundError

    if src.is_file():
        files = [src.as_posix()]
    elif src.is_dir():
        files = []
    else:
        debug(f"{src} is neither a file nor a directory")
        raise TypeError

    # Configure execution
    if not config:
        configpath = Path(__file__).parent
        configfile = configpath.joinpath("config.yml").as_posix()
        config = load_config(configfile, src=src)

    # Include files
    for incfile in config["include"]["files"]:
        files += set(src.glob(incfile))

    # Exclude files
    files = list(set(files) - set(config["exclude"]["files"]))

    # Scan files
    whispers = WhisperSecrets(config)
    for filename in files:
        for secret in whispers.scan(filename):
            if secret:
                yield secret



import json
import re
from collections import namedtuple
from difflib import SequenceMatcher
from hashlib import md5
from pathlib import Path
from typing import Optional

from yaml import safe_load

Secret = namedtuple("Secret", ["file", "line", "key", "value", "message", "severity"])


def strip_string(value: str) -> str:
    """
    Strips leading and trailing quotes and spaces
    """
    if not value:
        return ""
    return str(value).strip(" '\"\n\r\t")


def simple_string(value: str) -> str:
    """
    Returns a simplified value for loose comparison
    """
    if not value:
        return ""
    value = strip_string(value)  # Remove quotes
    value = value.rstrip("\\")  # Remove trailing backslashes
    value = value.lower()  # Lowercase
    value = re.sub(r"[^a-z0-9]", "_", value.strip())  # Simplify
    return value


def similar_strings(a: str, b: str) -> float:
    """
    Returns similarity coefficient between two strings
    """
    a = simple_string(a).replace("_", "")
    b = simple_string(b).replace("_", "")
    return SequenceMatcher(None, a, b).ratio()


def line_with_key_value(key: str, value: str, line: str) -> bool:
    """
    Both key and value are on the same line
    """
    return key in line and value in line


def line_with_value(value: str, line: str) -> bool:
    """
    Line containing just the value
    """
    return value in line


def line_begins_with_value(value: str, line: str) -> bool:
    """
    Find the line where the value begins
    """
    value_str = simple_string(value)
    line_str = simple_string(line)
    if not line_str:
        return False
    return value_str.startswith(line_str)


def find_line_number(filepath: Path, key: str, value: str) -> int:
    """
    Returns line number in file with given key and value
    """
    if not key:
        key = ""
    if not value:
        value = ""
    value = value.split("\n")[0]
    value_line_number = 0
    for line_number, line in enumerate(filepath.open().readlines(), 1):
        if not line:
            continue
        if line_with_key_value(key, value, line):
            return line_number
        elif line_begins_with_value(value, line):
            return line_number
        elif line_with_value(value, line):
            value_line_number = line_number
    return value_line_number


def load_yaml_from_file(filepath: Path) -> dict:
    ret = safe_load(filepath.read_text())
    if not isinstance(ret, dict):
        return {}
    return ret


def secret_checksum(secret: Secret) -> str:
    secret = json.dumps(secret._asdict())
    chk = md5()
    chk.update(secret.encode("utf-8"))
    return chk.hexdigest()


def format_secret(secret: Secret) -> str:
    return (
        secret_checksum(secret)
        + ":\n"
        + '  file: "'
        + str(secret.file)
        + '"\n'
        + '  line: "'
        + str(secret.line)
        + '"\n'
        + '  key: "'
        + str(secret.key)
        + '"\n'
        + '  value: "'
        + str(secret.value)
        + '"\n'
        + '  message: "'
        + str(secret.message)
        + '"\n'
        + '  severity: "'
        + str(secret.severity)
        + '"\n'
        + "\n"
    )


def format_stdout(secret: Secret, output: Optional[Path] = None) -> str:
    if output:
        output.open("a").write(format_secret(secret))
    data = json.dumps(secret._asdict())
    print(data)
    return data



from pathlib import Path
from typing import Optional

from whispers.plugins.config import Config
from whispers.plugins.dockerfile import Dockerfile
from whispers.plugins.htpasswd import Htpasswd
from whispers.plugins.jproperties import Jproperties
from whispers.plugins.json import Json
from whispers.plugins.npmrc import Npmrc
from whispers.plugins.pip import Pip
from whispers.plugins.plaintext import Plaintext
from whispers.plugins.pypirc import Pypirc
from whispers.plugins.python import Python
from whispers.plugins.shell import Shell
from whispers.plugins.xml import Xml
from whispers.plugins.yml import Yml


class WhisperPlugins:
    def __init__(self, filename: str):
        """Inits the rules objects. Call pairs() to get results"""
        self.filename = filename
        self.filepath = Path(filename)
        self.filetype = self.filepath.name.split(".")[-1]
        self.plugin = self.load_plugin()

    def load_plugin(self) -> Optional[object]:
        """Loads the correct plugin for a given file"""
        if not self.filepath.exists():
            return None
        elif not self.filepath.is_file():
            return None
        elif self.filepath.stat().st_size < 7:
            return None
        elif self.filetype in ["yaml", "yml"]:
            return Yml()
        elif self.filetype == "json":
            return Json()
        elif self.filetype == "xml":
            return Xml()
        elif self.filetype.startswith("npmrc"):
            return Npmrc()
        elif self.filetype.startswith("pypirc"):
            return Pypirc()
        elif self.filepath.name == "pip.conf":
            return Pip()
        elif self.filetype in ["conf", "cfg", "config", "ini", "env", "credentials"]:
            if self.filepath.open("r").readline().startswith("<?xml "):
                return Xml()
            else:
                return Config()
        elif self.filetype == "properties":
            return Jproperties()
        elif self.filetype.startswith(("sh", "bash", "zsh", "env")):
            return Shell()
        elif self.filepath.name.startswith("Dockerfile"):
            return Dockerfile()
        elif self.filetype.startswith("htpasswd"):
            return Htpasswd()
        elif self.filetype == "txt":
            return Plaintext()
        elif self.filetype == "py":
            return Python()
        return None

    def pairs(self):
        if self.plugin:
            yield from self.plugin.pairs(self.filepath)



from urllib.parse import parse_qsl, urlparse


class Uri:
    def pairs(self, code: str):
        uri = urlparse(code)
        if uri.password:
            yield "URI Password", uri.password
        if uri.query:
            for key, value in parse_qsl(uri.query):
                yield key, value



import shlex
from pathlib import Path

from whispers.utils import strip_string


class Shell:
    def pairs(self, filepath: Path):
        full_command = []
        for line in filepath.open("r").readlines():
            line = line.strip()
            if line.endswith("\\"):  # Multi-line commands
                full_command.append(line[:-1])
                continue
            full_command.append(line)
            cmd = shlex.split(" ".join(full_command))
            full_command = []
            if not cmd:
                continue
            elif cmd[0].lower() == "curl":
                yield from self.curl(cmd)
            for item in cmd:
                if "=" in item and len(item.split("=")) == 2:
                    key, value = item.split("=")
                    yield key, value

    def curl(self, cmd):
        indicators_combined = ["-u", "--user", "-U", "--proxy-user", "-E", "--cert"]
        indicators_single = ["--tlspassword", "--proxy-tlspassword"]
        indicators = indicators_combined + indicators_single
        for indicator in indicators:
            if indicator not in cmd:
                continue
            idx = cmd.index(indicator)
            if len(cmd) == idx + 1:
                continue  # End of command
            credentials = strip_string(cmd[idx + 1])
            if indicator in indicators_single:
                yield "cURL Password", credentials
            else:
                if ":" not in credentials:
                    continue  # Password not specified
                yield "cURL Password", credentials.split(":")[1]



from pathlib import Path

from whispers.utils import strip_string


class Htpasswd:
    def pairs(self, filepath: Path):
        for line in filepath.open("r").readlines():
            if ":" not in line:
                continue
            creds = line.split(":")
            value = strip_string(creds[1])
            if value:
                yield "htpasswd Hash", value



from pathlib import Path
from urllib.parse import urlparse


class Pip:
    def pairs(self, filepath: Path):
        for line in filepath.open("r").readlines():
            if "http" not in line:
                continue

            value = urlparse(line.split("=")[-1].strip()).password
            if value:
                yield "pip password", value



import pytest

from whispers import core

from .conftest import CONFIG_PATH, FIXTURE_PATH, config_path, fixture_path


@pytest.mark.parametrize("configfile", ["exclude_keys.yml", "exclude_values.yml"])
@pytest.mark.parametrize("src", ["excluded.yml", "excluded.json", "excluded.xml"])
def test_exclude_by_keys_and_values(configfile, src):
    config = core.load_config(config_path(configfile), FIXTURE_PATH)
    secrets = core.run(fixture_path(src), config=config)
    assert next(secrets).key == "hardcoded_password"
    with pytest.raises(StopIteration):
        next(secrets)


@pytest.mark.parametrize(
    ("src", "keys"),
    [
        ("privatekeys.yml", ["access", "key", "rsa", "dsa", "ec", "openssh"]),
        ("privatekeys.json", ["access", "key", "rsa", "dsa", "ec", "openssh"]),
        ("privatekeys.xml", ["access", "key", "rsa", "dsa", "ec", "openssh"]),
        ("aws.yml", ["aws_id", "aws_key", "aws_token"]),
        ("aws.json", ["aws_id", "aws_key", "aws_token"]),
        ("aws.xml", ["aws_id", "aws_key", "aws_token"]),
        ("jenkins.xml", ["noncompliantApiToken", "noncompliantPasswordHash"]),
        ("cloudformation.yml", ["NoncompliantDBPassword"]),
        ("cloudformation.json", ["NoncompliantDBPassword"]),
    ],
)
def test_detection_by_key(src, keys):
    secrets = core.run(fixture_path(src))
    for key in keys:
        assert next(secrets).key == key
    with pytest.raises(StopIteration):
        next(secrets)


@pytest.mark.parametrize(
    ("src", "count"),
    [
        ("custom.yml", 0),
        ("custom.json", 0),
        ("custom.xml", 0),
        ("hardcoded.yml", 5),
        ("hardcoded.json", 5),
        ("hardcoded.xml", 5),
        ("passwords.yml", 4),
        ("passwords.json", 4),
        ("passwords.xml", 4),
        ("placeholders.yml", 0),
        ("placeholders.json", 0),
        ("placeholders.xml", 0),
        ("apikeys.yml", 10),
        ("apikeys.json", 10),
        ("apikeys.xml", 10),
        (".npmrc", 3),
        (".pypirc", 1),
        ("pip.conf", 2),
        ("integration.conf", 5),
        ("integration.yml", 5),
        ("integration.json", 5),
        ("integration.xml", 5),
        ("settings.conf", 1),
        ("settings.cfg", 1),
        ("settings.ini", 1),
        ("settings.env", 1),
        ("shell.sh", 13),
        ("Dockerfile", 3),
        ("beans.xml", 3),
        ("jdbc.xml", 3),
        (".htpasswd", 2),
        (".aws/credentials", 3),
        ("falsepositive.yml", 4),
        ("source.py", 5),
        ("plaintext.txt", 2),
        ("uri.yml", 2),
        ("java.properties", 2),
    ],
)
def test_detection_by_value(src, count):
    config = core.load_config(CONFIG_PATH.joinpath("detection_by_value.yml"))
    secrets = core.run(fixture_path(src), config)
    for _ in range(count):
        value = next(secrets).value.lower()
        if value.isnumeric():
            value = bytes.fromhex(hex(int(value))[2:]).decode("ascii")
        assert "hardcoded" in value
    with pytest.raises(StopIteration):
        next(secrets)


def test_detection_by_filename():
    expected = map(
        fixture_path,
        [
            ".aws/credentials",
            ".htpasswd",
            ".npmrc",
            ".pypirc",
            "connection.config",
            "integration.conf",
            "pip.conf",
            "settings.cfg",
            "settings.conf",
            "settings.env",
            "settings.ini",
        ],
    )
    config = core.load_config(CONFIG_PATH.joinpath("detection_by_filename.yml"))
    secrets = core.run(fixture_path(""), config)
    result = [secret.value for secret in secrets]
    for exp in expected:
        assert exp in result



import os
from importlib import import_module
from pathlib import Path

from setuptools import find_packages, setup


def get_version():
    return import_module("whispers.__version__").__version__


install_requires = ["lxml>=4.5.0", "pyyaml>=5.3", "astroid>=2.3.3", "jproperties==2.1.0"]

dev_requires = [
    "black>=19.10b0",
    "coverage~=4.4",
    "flake8>=3.8.1",
    "isort>=4.3.21",
    "pytest>=2.9.1",
    "pytest-mock>=1.0",
    "pip-tools>=4.4.1",
    "wheel>=0.34.2",
    "twine>=3.1.1",
]

setup(
    name="whispers",
    version=get_version(),
    url="https://github.com/Skyscanner/whispers",
    author="Artëm Tsvetkov",
    author_email="artem.tsvetkov@skyscanner.net",
    description="Identify secrets and dangerous behaviours",
    packages=find_packages(),
    include_package_data=True,
    platforms="any",
    install_requires=install_requires,
    setup_requires=["pytest-runner"],
    tests_require=dev_requires,
    extras_require={"dev": dev_requires},
    entry_points={"console_scripts": ["whispers=whispers.cli:cli"]},
)



import logging
import logging.config
import traceback
from pathlib import Path


def configure_log(path: str = ""):
    path = Path(path, "whispers.log")
    path.write_text("")
    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "formatters": {
                "default": {
                    "class": "logging.Formatter",
                    "style": "{",
                    "datefmt": "%Y-%m-%d %H:%M",
                    "format": "[{asctime:s}] {message:s}\n",
                }
            },
            "handlers": {
                "default": {
                    "level": "DEBUG",
                    "class": "logging.handlers.WatchedFileHandler",
                    "formatter": "default",
                    "filename": path.as_posix(),
                    "mode": "w",
                    "encoding": "utf-8",
                }
            },
            "loggers": {"": {"handlers": ["default"], "level": "DEBUG", "propagate": False}},  # root logger
        }
    )


def debug(message: str = "") -> str:
    trace = traceback.format_exc().strip()
    if trace == "NoneType: None":
        ret = message
    else:
        ret = f"{trace}\n{message}"
    log = logging.getLogger()
    log.warning(ret)
    return ret



from whispers.plugins.uri import Uri
from whispers.rules import WhisperRules


class StructuredDocument:
    def __init__(self):
        self.breadcrumbs = []
        self.rules = WhisperRules()

    def traverse(self, code, key=None):
        """Recursively traverse YAML/JSON document"""
        if isinstance(code, dict):
            yield from self.cloudformation(code)
            for k, v in code.items():
                self.breadcrumbs.append(k)
                yield k, v, self.breadcrumbs
                yield from self.traverse(v, key=k)
                self.breadcrumbs.pop()
            # Special key/value format
            elements = list(code.keys())
            if "key" in elements and "value" in elements:
                yield code["key"], code["value"], self.breadcrumbs
        elif isinstance(code, list):
            for item in code:
                yield key, item, self.breadcrumbs
                yield from self.traverse(item, key=key)
        elif isinstance(code, str):
            if "=" in code:
                item = code.split("=", 1)
                if len(item) == 2:
                    yield item[0], item[1], self.breadcrumbs
            if self.rules.match("uri", code):
                for k, v in Uri().pairs(code):
                    yield k, v, self.breadcrumbs

    def cloudformation(self, code):
        """
        AWS CloudFormation format
        """
        if self.breadcrumbs:
            return  # Not tree root
        if "AWSTemplateFormatVersion" not in code:
            return  # Not CF format
        if "Parameters" not in code:
            return  # No parameters
        for key, values in code["Parameters"].items():
            if "Default" not in values:
                continue  # No default value
            yield key, values["Default"]
